import typing
from functools import lru_cache
from urllib import parse
import textwrap

from wikitools import console, reference_parser

_urlparse = lru_cache(maxsize=4096)(parse.urlparse)


# Not an enum/class to avoid attribute lookups
_STATE_IDLE = 0  # IDLE: jump to next '[' using optimized str.find
_STATE_START = 1  # START: tracking bracket depth inline
_STATE_INLINE = 2  # INLINE: tracking paren depth inline
_STATE_REFERENCE = 3  # REFERENCE: tracking bracket depth inline


def _shorten(s):
    return textwrap.shorten(s, 20, placeholder="...")


class Link(typing.NamedTuple):
    """
    A Markdown link, inline- or reference-style, external or internal.
    May be relative. Example:

        See [Difficulty Names](/wiki/Beatmap/Difficulty#naming-conventions)

    - alt_text: 'Difficulty Names'
    - raw_location: '/wiki/Beatmap/Difficulty#naming-conventions'
    - parsed_location: urllib.parse.ParseResult with all of its fields

    Another example:

        ![Player is AFK](img/chat-console-afk.png "Player is away from keyboard")

    - alt_text: 'Player is AFK'
    - raw_location: 'img/chat-console-afk.png'
    - parsed_location: urllib.parse.ParseResult with all of its fields
    - title: ' "Player is away from keyboard"'
    """

    # Link position within the line. Example:
    #   See also: [Difficulty names](/wiki/Beatmap/Difficulty#naming-conventions)
    #             ^ start                                                       ^ end
    start: int
    end: int

    # Sections of a link. Example:
    #    ![Player is AFK](img/chat-console-afk.png "Player is away from keyboard")
    #      ^ alt_text  ^
    #                     ^ ----- location ----- ^
    #                                             ^ ---------- title ---------- ^
    #                     ^ --------------------- content --------------------- ^
    #     ^ ------------------ full_link / full_coloured_link ------------------ ^
    alt_text: str
    raw_location: str
    parsed_location: parse.ParseResult
    title: str

    @property
    def content(self):
        return self.raw_location if not self.title else f"{self.raw_location} {self.title}"

    @property
    def full_link(self):
        if self.is_reference:
            return f"[{self.alt_text}][{self.content}]"
        else:
            return f"[{self.alt_text}]({self.content})"

    @property
    def truncated_coloured_link(self, fragment_only=False):
        return "{alt_text_in_braces}{left_brace}{location}{extra}{right_brace}".format(
            alt_text_in_braces=console.green(f"[{_shorten(self.alt_text)}]"),
            left_brace=console.green('[') if self.is_reference else console.green('('),
            location=self.colourise_location(),
            extra=" " + console.blue("\"...\"") if self.title else "",
            right_brace=console.green(']') if self.is_reference else console.green(')'),
        )

    @property
    def truncated_coloured_link_fragment(self, fragment_only=False):
        return "{alt_text_in_braces}{left_brace}{location}{extra}{right_brace}".format(
            alt_text_in_braces=console.green(f"[{_shorten(self.alt_text)}]"),
            left_brace=console.green('[') if self.is_reference else console.green('('),
            location=self.colourise_location(fragment_only=True),
            extra=" " + console.blue("\"...\"") if self.title else "",
            right_brace=console.green(']') if self.is_reference else console.green(')'),
        )

    @property
    def fragment_start(self):
        """
        Position of the link #fragment in the line, if there is one. Otherwise, the same value as the end of the link.
        """
        return self.start + len(self.alt_text) + 2 + len(self.parsed_location.path) + 1

    def colourise_link(self, fragment_only=False):
        return "{alt_text_in_braces}{left_brace}{location}{extra}{right_brace}".format(
            alt_text_in_braces=console.green(f"[{self.alt_text}]"),
            left_brace=console.green('[') if self.is_reference else console.green('('),
            location=self.colourise_location(fragment_only=fragment_only),
            extra=" " + console.blue(self.title) if self.title else "",
            right_brace=console.green(']') if self.is_reference else console.green(')'),
        )

    def colourise_location(self, fragment_only=False):
        return self.colourise_location_static(self.raw_location.split("#")[0], self.parsed_location.fragment, fragment_only=fragment_only)

    # provided for convenience, used in `BrokenRedirectError`
    @staticmethod
    def colourise_location_static(location: str, fragment: typing.Optional[str] = None, fragment_only: bool = False):
        if fragment_only:
            colourised_location = console.green(location)
            if fragment:
                colourised_location += console.red('#' + fragment)
            return colourised_location
        return console.red(location + ('#' + fragment if fragment else ""))

    def resolve(
        self, references: reference_parser.References
    ) -> typing.Optional[reference_parser.Reference]:
        if not self.is_reference:
            return None
        return references.get(self.parsed_location.path)

    # Whether the link is a reference-style link. The only difference is that
    # `location` is a reference and needs to be resolved later.
    #
    # The syntax for such links is the almost the same as regular links:
    #    [text][reference]
    #
    # The reference can then later be defined at the start of a new line:
    #    [reference]: link
    is_reference: bool


def find_link(s: str, index=0) -> typing.Optional[Link]:
    """
    Finds the first valid Markdown link found in the string `s`, starting the search from position `index`.
    The following are considered links (title and alt text may be omitted):
        - [alt_text](/loca/ti/on "Title")
        - ![alt_text](/path/to/image "Title), with ! not being considered a part of the link
        - [artist - title (creator) [diff]](/loca/ti/ion_(with_parentheses) "Title")
        - [alt_text][reference], with exact locations found separately via find_reference()
    """

    state = _STATE_IDLE

    start = -1
    location = -1
    extra = None

    in_leading_whitespace = True

    bracket_depth = 0
    paren_depth = 0

    s_len = len(s)
    i = index

    while i < s_len:
        if state == _STATE_IDLE:
            i = s.find('[', i)
            if i == -1:
                return None

            # potential start of a link
            bracket_depth = 1
            state = _STATE_START
            start = i
            i += 1
            continue

        c = s[i]

        if state == _STATE_START:
            if c == '[':
                bracket_depth += 1
            elif c == ']':
                bracket_depth -= 1
            if bracket_depth == 0:
                if s_len <= i + 1:
                    # end of the line
                    state = _STATE_IDLE
                    i += 1
                    continue

                if s[start + 1] == '^':
                    # found a footnote -> ignore
                    state = _STATE_IDLE
                    i += 1
                    continue

                # the end of a bracket. the link may continue
                # to be inline- or reference-style
                if s[i + 1] == '(':
                    state = _STATE_INLINE
                    location = i + 2
                    paren_depth = 0
                elif s[i + 1] == '[':
                    if i + 2 < s_len and s[i + 2] == '^':
                        # found a footnote after bracket pair -> ignore
                        state = _STATE_IDLE
                        i += 1
                        continue

                    state = _STATE_REFERENCE
                    location = i + 2
                    bracket_depth = 0
                else:
                    state = _STATE_IDLE
            i += 1
            continue

        if state == _STATE_INLINE:
            if c == ' ' and not in_leading_whitespace and extra is None:
                extra = i
            elif c != ' ' and in_leading_whitespace and paren_depth > 0:
                in_leading_whitespace = False

            if c == '(':
                paren_depth += 1
            elif c == ')':
                paren_depth -= 1
            if paren_depth == 0:
                # end of a complete link
                if extra is None:
                    extra = i

                raw_location = s[location: extra]
                return Link(
                    raw_location=raw_location,
                    parsed_location=_urlparse(raw_location.strip()),
                    alt_text=s[start + 1: location - 2],
                    title=s[extra: i],
                    start=start,
                    end=i,
                    is_reference=False
                )
            i += 1
            continue

        if state == _STATE_REFERENCE:
            if c == '[':
                bracket_depth += 1
            elif c == ']':
                bracket_depth -= 1
            if bracket_depth == 0:
                # end of a complete reference-style link
                raw_location = s[location: i]
                return Link(
                    raw_location=raw_location,
                    parsed_location=_urlparse(raw_location),
                    alt_text=s[start + 1: location - 2],
                    title="",
                    start=start,
                    end=i,
                    is_reference=True
                )
            i += 1
            continue

        i += 1

    return None


def find_links(line: str) -> typing.List[Link]:
    """
    Iteratively extract all links from a line.
    """

    if '[' not in line:
        return []

    results = []
    match = find_link(line)
    while match:
        results.append(match)
        match = find_link(line, match.end + 1)
    return results



#!/usr/bin/env python3

import argparse
import os
import sys
import typing
import json
from itertools import chain

import yamllint.cli  # type: ignore
import yamllint.config  # type: ignore
import yamllint.linter  # type: ignore
import yamllint.rules  # type: ignore
from yamllint.linter import LintProblem

from wikitools import console
from wikitools import yaml_rules

FRONT_MATTER_DELIMITER = "---"
MARKDOWN_EXTENSION = ".md"

DEFAULT_CONFIG_CONTENT = 'extends: default'


def print_clean():
    print(f"{console.grey('Notice:')} No errors in YAML files detected.", file=sys.stderr)


def rewrite_problem(problem: LintProblem, path: str) -> LintProblem:
    """
    Rewrites a yamllint error to be less confusing and more consistent with other osu-wiki-tools errors
    """

    # yamllint: yamllint/linter.py:180
    if problem.desc.startswith("syntax error: ") and problem.desc.endswith(" (syntax)"):
        problem.desc = problem.desc[14:-9]
        problem.rule = "syntax"

        # pyyaml: lib/yaml/scanner.py
        if problem.desc == "could not find expected ':'" and path == "wiki/redirect.yaml":
            problem.desc += " between redirect and its destination"

    if not problem.rule:
        problem.rule = "unknown"

    return problem


def install_custom_checks(config: yamllint.config.YamlLintConfig):
    for rule_cls in yaml_rules.OSU_WIKI_RULES:
        config.rules[rule_cls.ID] = yamllint.rules._RULES[rule_cls.ID] = rule_cls()


def front_matter(fileobj: typing.TextIO, filepath: str) -> str:
    delimiters = 0
    lines = []
    for line in fileobj:
        if line.split("#")[0].strip() == FRONT_MATTER_DELIMITER:
            delimiters += 1
        # Stop on second delimiter, or when it's clear there won't be front matter at all
        if delimiters == 2 or line.startswith("# "):
            break
        lines.append(line)

    if delimiters == 0:
        return ""
    if delimiters == 2:
        return "".join(lines)
    raise ValueError(f"The file {filepath} has malformed front matter")


def read_yaml(filepath: str):
    with open(filepath, "r", encoding='utf-8') as fd:
        if filepath.endswith(MARKDOWN_EXTENSION):
            return front_matter(fd, filepath)
        else:
            return fd.read()


def file_iterator(roots: list, config: yamllint.config.YamlLintConfig):
    for item in roots:
        if os.path.isdir(item):
            for root, _, filenames in os.walk(item):
                for f in filenames:
                    filepath = os.path.join(root, f)
                    if (
                        not config.is_file_ignored(filepath) and
                        (config.is_yaml_file(filepath) or filepath.endswith(MARKDOWN_EXTENSION))
                    ):
                        yield filepath
        elif os.path.isfile(item):
            yield item


def errors_json(errors: typing.List[typing.Tuple[str, LintProblem]]) -> str:
    return json.dumps(
        [
            {
                "level": error.level,
                "path": path,
                "line": error.line,
                "column": error.column,
                "type": "yamllint:" + error.rule,
                "text": error.desc,
            }
            for path, error in errors
        ]
    )


def parse_args(args: list):
    parser = argparse.ArgumentParser(
        description="Check YAML files and front matter of the Markdown documents for common mistakes"
    )
    parser.add_argument("--config", help="path to the yamllint config", default="./.yamllint.yaml")
    parser.add_argument(
        "--target", nargs="+", default=(".",),
        help="files or directories to lint (by default, assumes current working directory)"
    )
    parser.add_argument(
        "--format", choices=("parsable", "standard", "colored", "github", "auto", "json"),
        default="colored", help="output format (see yamllint docs for details)"
    )
    return parser.parse_args(args)


def main(*args):
    args = parse_args(args)
    if os.path.exists(args.config):
        config = yamllint.config.YamlLintConfig(file=os.path.expanduser(args.config))
    else:
        config = yamllint.config.YamlLintConfig(DEFAULT_CONFIG_CONTENT)

    install_custom_checks(config)

    exit_code = 0
    warnings = []
    errors = []
    problems_grouped = []

    for path in file_iterator(args.target, config):
        try:
            path = path[2:] if path.startswith('./') else path
            payload = read_yaml(path)
            problems = yamllint.linter.run(payload, config, path)
        except EnvironmentError as e:
            print(e, file=sys.stderr)
            return -1

        problems = [rewrite_problem(problem, path) for problem in problems]
        problems_grouped.append((path, problems))

        for problem in problems:
            if problem.level == "error":
                errors.append((path, problem))
                exit_code = 1
            elif problem.level == "warning":
                warnings.append((path, problem))

    if exit_code == 0:
        print_clean()
        return exit_code

    match args.format:
        case "github":
            print("::group::Annotations")
            for path, error in errors[:10]:
                error = rewrite_problem(error, path)
                print("::error file={},line={},col={},title={}::{}".format(
                    path,
                    error.line,
                    error.column,
                    "yamllint:" + error.rule,
                    error.desc,
                ))
            print("::endgroup::\n")

            for path, problems in problems_grouped:
                yamllint.cli.show_problems(problems, path, args_format="colored", no_warn=False)

        case "json":
            print(errors_json(chain(errors, warnings)))

        case _:
            for path, problems in problems_grouped:
                yamllint.cli.show_problems(problems, path, args_format=args.format, no_warn=False)

    return exit_code


if __name__ == '__main__':
    sys.exit(main(*sys.argv[1:]))



import argparse
import json
import sys
import typing

from wikitools.article_parser import load_front_matter, save_front_matter


class KeyValue(typing.NamedTuple):
    key: str
    value: str


class KeyValueAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        try:
            items = [KeyValue(*(s.strip() for s in v.split(":", 1))) for v in values]
        except TypeError:
            parser.error(f"{option_string or self.dest}: value must be of format key:value")
        setattr(namespace, self.dest, items)


def parse_args(args):
    parser = argparse.ArgumentParser(usage="%(prog)s front-matter [options]")
    parser.add_argument("file", type=argparse.FileType("r", encoding="utf-8"), help="Markdown file to edit")
    parser.add_argument("-p", "--print", nargs="*", help="front matter items to print in JSON (or everything by default)")
    parser.add_argument("-s", "--set", nargs="+", action=KeyValueAction, metavar="KEY:VALUE", help="front matter items to add or edit")
    parser.add_argument("-r", "--remove", nargs="+", metavar="KEY", help="front matter items to remove if they exist")
    return parser.parse_args(args)


def main(*args):
    args = parse_args(args)
    old_front_matter = load_front_matter(args.file)
    front_matter = dict(old_front_matter)

    if args.print is not None:
        to_print = {}
        for item in args.print or front_matter.keys():
            try:
                to_print[item] = front_matter[item]
            except KeyError:
                to_print[item] = None
        print(json.dumps(to_print, indent=4))

    if args.set:
        for item in args.set:
            front_matter[item.key] = item.value

    if args.remove:
        for item in args.remove:
            del front_matter[item]

    if front_matter != old_front_matter:
        save_front_matter(args.file.name, front_matter)

    return 0


if __name__ == "__main__":
    sys.exit(main(*sys.argv[1:]))



#!/usr/bin/env python3

import argparse
import sys
import typing
from pathlib import Path
import json

from wikitools import console, errors as error_types, file_utils
from wikitools.file_utils import exists_case_sensitive


def print_error(error: error_types.FileError):
    print(error.pretty_location())
    print(error.pretty())
    print()


def print_clean():
    print(f"{console.grey('Notice:')} No file or folder structure errors detected.", file=sys.stderr)


def s(i: int, s: str) -> str:
    return f"{i} {s}{'s' if i != 1 else ''}"


def print_count(errors: int, files: int):
    print(f"{console.blue('Note:')} Found {s(errors, 'error')} ({s(files, 'file')} checked).")


def check_missing_english_version(file_path: Path) -> typing.Optional[error_types.MissingEnglishVersionError]:
    path = Path(file_path)
    dir_name = path.parent
    english_path = dir_name / "en.md"
    if not exists_case_sensitive(english_path):
        return error_types.MissingEnglishVersionError(file_path)

    return None


def files_deduplicated(file_paths: typing.List[Path]) -> typing.List[Path]:
    seen = set()
    deduplicated = []
    for file in sorted(file_paths):
        dir = file.parent
        if dir not in seen:
            deduplicated.append(file)
            seen.add(dir)
    return deduplicated


def errors_json(errors: typing.List[error_types.FileError]) -> str:
    return json.dumps(
        [
            {
                "path": error.path.as_posix(),
                "type": "file-checker:" + error.id,
                "text": repr(error),
            }
            for error in errors
        ]
    )


def parse_args(args):
    parser = argparse.ArgumentParser(usage="%(prog)s check-files [options]")
    parser.add_argument("-t", "--target", nargs='*', help="paths to the articles you want to check, relative to the repository root")
    parser.add_argument("-a", "--all", action='store_true', help="check all articles")
    parser.add_argument("-f", "--format", choices=["regular", "json", "github"], default="regular", help="specify output format")
    parser.add_argument("-r", "--root", help="specify repository root, current working directory assumed otherwise")
    return parser.parse_args(args)


def main(*args):
    args = parse_args(args)
    if not args.target and not args.all:
        print(f"{console.grey('Notice:')} No articles to check.", file=sys.stderr)
        sys.exit(0)

    if args.root:
        changed_cwd = file_utils.ChangeDirectory(args.root)  # Keep alive to maintain directory change  # noqa: F841

    if args.all:
        filenames = file_utils.list_all_articles()
    else:
        filenames = list(filter(lambda x: file_utils.is_article(x) or file_utils.is_newspost(x), args.target))

    exit_code = 0
    error_count = 0
    file_count = 0
    all_errors = []

    for filename in files_deduplicated([Path(f) for f in filenames]):
        file_count += 1

        error = check_missing_english_version(filename)

        if error:
            exit_code = 1
            error_count += 1
            all_errors.append(error)

    if exit_code == 0:
        print_clean()
        return exit_code

    match args.format:
        case "regular":
            for error in all_errors:
                print_error(error)

            print_count(error_count, file_count)

        case "json":
            print(errors_json(all_errors))

        case "github":
            print("::group::Annotations")
            for error in all_errors[:10]:
                print(f"::error file={error.file},title={"file-checker:" + error.id}::{repr(error)}")
            print("::endgroup::\n")

            for error in all_errors:
                print_error(error)

            print_count(error_count, file_count)

    if args.root:
        del changed_cwd
    return exit_code


if __name__ == "__main__":
    sys.exit(main(*sys.argv[1:]))




import tests.utils as utils

import pytest
import textwrap

from wikitools_cli.commands import check_yaml as yaml_checker


class TestCheckFiles:
    @pytest.mark.parametrize(
        "format",
        [
            "parsable",
            "standard",
            "colored",
            "github",
            "auto",
            "json"
        ]
    )
    @pytest.mark.parametrize(
        "payload",
        [
            {
                "path": 'wiki/Article/en.md',
                "content": textwrap.dedent('''
                    ---
                    stub: true
                    tags:
                      - tag
                    ---

                    # An article
                ''').strip()
            },
            {
                "path": 'wiki/redirect.yaml',
                "content": textwrap.dedent('''
                    # yamllint disable rule:colons

                    asc:                           Article_styling_criteria
                    asg:                           Article_styling_criteria
                    asc/images:                    Article_styling_criteria#images
                    asg/images:                    Article_styling_criteria#images
                    wiki_styling_criteria:         Article_styling_criteria
                    wsc:                           Article_styling_criteria
                ''').strip() + "\n"
            },
        ]
    )
    def test__check_yaml_valid(self, root, format, payload):
        utils.create_files(root, (payload["path"], payload["content"]))

        exit_code = yaml_checker.main("--format", format)
        assert exit_code == 0

    @pytest.mark.parametrize(
        "format",
        [
            "parsable",
            "standard",
            "colored",
            "github",
            "auto",
            "json"
        ]
    )
    @pytest.mark.parametrize(
        "payload",
        [
            {
                "path": 'wiki/Article/en.md',
                "content": textwrap.dedent('''
                    ---
                    stub: true
                    unknown_tag: true
                    tags:
                        - outdated: true
                    ---

                    # An article
                ''').strip()
            },
            {
                "path": 'wiki/redirect.yaml',
                "content": textwrap.dedent('''
                    # definitely not valid yaml all around

                    "asc": "Article_Styling_Criteria"
                    "asc/images": "Article_Styling_Criteria#images"

                    "ignore_list": "Client/Options/Ignore_list"
                    "ignore":      "Client/Options/Ignore_list"
                    unquoted_key1:  unquoted/value1
                    "quoted_key": unquoted/value2
                    unquoted_key2:    "quoted/value"
                    osu!:    "Disambiguation/osu!"
                    osu!:rules:    "Rules"
                ''').strip()
            },
        ]
    )
    def test__check_yaml_invalid(self, root, format, payload):
        utils.create_files(root, (payload["path"], payload["content"]))

        exit_code = yaml_checker.main("--format", format)
        assert exit_code == 1



import json

import pytest
import tests.utils as utils
from tests.utils import OutputCapture

from wikitools_cli.commands import front_matter


class TestEditFrontMatter:
    @pytest.mark.parametrize(
        "payload",
        [
            {
                "path": "wiki/None/en.md",
                "content": "# Article",
                "args": ("--set", "layout:post"),
                "expected": "---\nlayout: post\n---\n\n# Article"
            },
            {
                "path": "wiki/Has/en.md",
                "content": "---\ntitle: test\n---\n\n# Article",
                "args": ("--set", "layout:post"),
                "expected": "---\ntitle: test\nlayout: post\n---\n\n# Article"
            },
            {
                "path": "wiki/Existing/en.md",
                "content": "---\ntitle: bad\n---\n\n# Article",
                "args": ("--set", "title:good"),
                "expected": "---\ntitle: good\n---\n\n# Article"
            },
            {
                "path": "wiki/Quoted/en.md",
                "content": "---\ntitle: unquoted\n---\n\n# Article",
                "args": ("--set", "title:unquoted 2: electric boogaloo"),
                "expected": "---\ntitle: \"unquoted 2: electric boogaloo\"\n---\n\n# Article"
            },
            {
                "path": "wiki/Remove/en.md",
                "content": "---\ntag: bye\n---\n\n# Article",
                "args": ("--remove", "tag"),
                "expected": "# Article"
            },
            {
                "path": "wiki/Remove_multiple/en.md",
                "content": "---\ntag: bye\nseries: yep\n---\n\n# Article",
                "args": ("--remove", "tag", "series"),
                "expected": "# Article"
            },
        ]
    )
    def test__edit_front_matter(self, root, payload):
        utils.create_files(root, (payload["path"], payload["content"]))

        exit_code = front_matter.main(payload["path"], *payload["args"])
        assert exit_code == 0

        with open(payload["path"], "r", encoding="utf-8") as file:
            content = file.read()

        assert content == payload["expected"]

    @pytest.mark.parametrize(
        "payload",
        [
            {
                "path": "wiki/None/en.md",
                "content": "# Article",
                "args": ("--print", "title"),
                "expected": {"title": None}
            },
            {
                "path": "wiki/Some/en.md",
                "content": "---\ntitle: test\n---\n\n# Article",
                "args": ("--print", "title"),
                "expected": {"title": "test"}
            },
            {
                "path": "wiki/Multiple/en.md",
                "content": "---\ntitle: cookies\nseries: cooking\n---\n\n# Article",
                "args": ("--print", "title", "series"),
                "expected": {"title": "cookies", "series": "cooking"}
            },
            {
                "path": "wiki/Partial/en.md",
                "content": "---\ntitle: cookies\n---\n\n# Article",
                "args": ("--print", "title", "series"),
                "expected": {"title": "cookies", "series": None}
            },
            {
                "path": "wiki/All/en.md",
                "content": "---\ntitle: cookies\nseries: cooking\n---\n\n# Article",
                "args": ("--print",),
                "expected": {"title": "cookies", "series": "cooking"}
            },
        ]
    )
    def test_print_front_matter(self, root, payload):
        utils.create_files(root, (payload["path"], payload["content"]))

        with OutputCapture() as out:
            exit_code = front_matter.main(payload["path"], *payload["args"])

        assert exit_code == 0
        assert out.stderr == ""
        assert json.loads(out.stdout) == payload["expected"]
